/**
 * 
 */
package com.btc.ep.dal.persist.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.persistence.EntityManager;
import javax.persistence.EntityTransaction;
import javax.persistence.FlushModeType;
import javax.persistence.Query;
import javax.persistence.TypedQuery;

import com.btc.ep.base.bl.ModelElement;

/**
 * JPA Generic/Base Data Access Object
 * 
 * @author Jacek
 *
 */
public class PersistenceService {

    /* *****************************************************
     * * Generics
     * ****************************************************
     */

    protected EntityManager entityManager;

    /**
     * create new Instance with entityManager
     * 
     * @param entityManager The EntityMangager
     */
    public PersistenceService(EntityManager entityManager) {
        this.entityManager = entityManager;
        entityManager.setFlushMode(FlushModeType.AUTO);
        EntityTransaction transaction = entityManager.getTransaction();
        transaction.begin();
    }

    /**
     * close the active entityManager
     *
     */
    public void close() {
        EntityTransaction transaction = entityManager.getTransaction();
        try {
            transaction.rollback();
        } finally {
            this.entityManager.close();
        }
    }

    /**
     * clear the first level cache of the entitymanager persistence context
     */
    public void clear() {
        this.entityManager.clear();
    }

    /**
     * add/persist new object
     * 
     * @param obj Entity to persist
     * @param <T> Type definition
     * @return the obj
     */
    @Deprecated
    public <T extends ModelElement> T create(T obj) {
        persist(obj);
        commit();
        return obj;
    }

    /**
     * add/persist new object
     * 
     * @param obj Entity to persist
     * @param <T> Type definition
     */
    public <T extends ModelElement> void persist(T obj) {
        EntityManager em = getEntityManager();
        em.persist(obj);
    }

    /**
     * merge changes
     * 
     * @param obj Entity to merge
     * @param <T> Type definition
     * @return the new and attached object
     */
    public <T extends ModelElement> T update(T obj) {
        EntityManager em = getEntityManager();
        T aobj = em.merge(obj);
        return aobj;
    }

    /**
     * refresh an entity
     * 
     * @param obj Entity to refresh
     * @param <T> Type definition
     */
    public <T extends ModelElement> void refresh(T obj) {
        EntityManager em = getEntityManager();
        em.refresh(obj);
    }

    /**
     * commit changes
     * 
     */
    public void commit() {
        EntityManager em = getEntityManager();
        EntityTransaction transaction = em.getTransaction();
        try {
            transaction.commit();
        } catch (Exception ex) {
            throw ex;
        } finally {
            transaction.begin();
        }
    }

    /**
     * commit changes
     * 
     */
    public void rollback() {
        EntityManager em = getEntityManager();
        EntityTransaction transaction = em.getTransaction();
        transaction.rollback();
        transaction.begin();
    }

    /**
     * flush changes
     * 
     */
    public void flush() {
        EntityManager em = getEntityManager();
        em.flush();
    }
    
    public EntityTransaction getTransaction() {
        
        EntityTransaction transaction = getEntityManager().getTransaction();
        
        if (transaction.isActive()) {
            return null;
        } else {
            transaction.begin();
            return transaction;
        }
    }


    /**
     * remove persisted object by id
     * 
     * @param clazz Type of the object
     * @param <T> Type definition
     * @param uid UUID of the Entity to be removed
     */
    public <T extends ModelElement> void remove(Class<T> clazz, String uid) {
        EntityManager em = getEntityManager();
        T obj = em.find(clazz, uid);
        em.remove(obj);
    }
    
    /**
     * remove given obj
     * 
     * @param obj Object to be removed
     * @param <T> Type definition
     * 
     * @param obj
     */
    public <T extends ModelElement> void remove(T obj) {
        EntityManager em = getEntityManager();
        em.remove(obj);
    }

    /**
     * check if persistenceContext contains object
     *
     * @param obj to be checked
     * @param <T> Type definition
     * 
     * @return true if persistenceContext contains obj
     */
    public <T extends ModelElement> boolean contains(T obj) {
        EntityManager em = getEntityManager();
        boolean res = em.contains(obj);
        return res;
    }

    /**
     * get/lookup object by id
     * 
     * @param clazz Type of the results
     * @param <T> Type definition
     * @param uid UUID of the Entity
     * @return The Entity Object
     */
    public <T extends ModelElement> T get(Class<T> clazz, Object uid) {
        EntityManager em = getEntityManager();
        return em.find(clazz, uid);
    }

    /**
     * find entity by primary key
     *
     * @param <T> class type inherited from ModelElement
     * @param clazz type of the result
     * @param uid primary key
     * @return ModelElement single Result
     */
    public <T extends ModelElement> T find(Class<T> clazz, String uid) {

        Map<String, Object> map = new HashMap<>();
        map.put("objectdb.result-fetch", "LAZY");
        EntityManager em = getEntityManager();
        return em.find(clazz, uid, map);
    }

    /**
     * find entity by primary key
     *
     * @param <T> class type inherited from ModelElement
     * @param clazz type of the result
     * @param uid primary key
     * @return ModelElement single Result
     */
    public <T extends ModelElement> T getReference(Class<T> clazz, String uid) {

        Map<String, Object> map = new HashMap<>();
        map.put("objectdb.result-fetch", "LAZY");
        EntityManager em = getEntityManager();
        return em.getReference(clazz, uid);
    }

    /**
     * get/lookup object by named query
     * 
     * @param query query id of the named query
     * @param clazz Type of the results
     * @param <T> Type definition
     * @param args query arguments
     * @return ModelElement JPQL single Result
     */
    public <T extends ModelElement> T getModelElementByNamedQuery(String query, Class<T> clazz, Object... args) {
        EntityManager em = getEntityManager();
        TypedQuery<T> q = em.createNamedQuery(query, clazz);
        for (int i = 0; i < args.length; i++) {
            Object arg = args[i];
            if (arg instanceof String) {
                arg = asLike((String)arg);
            }
            q.setParameter(i + 1, arg);
        }
        return q.getSingleResult();
    }

    /**
     * get/lookup objects by named query
     * 
     * @param query query id of the named query
     * @param clazz Type of the results
     * @param <T> Type definition
     * @param args query arguments
     * @return List<ModelElement> JPQL Result list
     */
    public <T extends ModelElement> List<T> getModelElementsByNamedQuery(String query, Class<T> clazz,
        Object... args) {
        EntityManager em = getEntityManager();
        TypedQuery<T> q = em.createNamedQuery(query, clazz);
        q.setFlushMode(FlushModeType.AUTO);
        for (int i = 0; i < args.length; i++) {
            Object arg = args[i];
            if (arg instanceof String) {
                arg = asLike((String)arg);
            }
            q.setParameter(i + 1, arg);
        }
        return q.getResultList();
    }

    /**
     * get/lookup ModelElement object by query
     * 
     * @param query JPQL query
     * @param clazz Type of the results
     * @param <T> Type definition
     * @param args query arguments
     * @return ModelElement JPQL single Result
     */
    public <T extends ModelElement> T getModelElementByQuery(String query, Class<T> clazz, Object... args) {
        EntityManager em = getEntityManager();
        TypedQuery<T> q = em.createQuery(query, clazz);
        for (int i = 0; i < args.length; i++) {
            Object arg = args[i];
            if (arg instanceof String) {
                arg = asLike((String)arg);
            }
            q.setParameter(i + 1, arg);
        }
        return q.getSingleResult();
    }

    /**
     * get/lookup ModelElement objects by query
     * 
     * @param query JPQL query
     * @param clazz Type of the results
     * @param <T> Type definition
     * @param args query arguments
     * @return List<ModelElement> JPQL Result list
     */
    public <T extends ModelElement> List<T> getModelElementsByQuery(String query, Class<T> clazz, Object... args) {
        EntityManager em = getEntityManager();
        TypedQuery<T> q = em.createQuery(query, clazz);
        q.setHint("objectdb.result-fetch", "LAZY");
        for (int i = 0; i < args.length; i++) {
            Object arg = args[i];
            if (arg instanceof String) {
                arg = asLike((String)arg);
            }
            q.setParameter(i + 1, arg);
        }
        return q.getResultList();
    }

    /**
     * get/lookup object by query
     * 
     * @param queryName query id
     * @param clazz Type of the results
     * @param <T> Type definition
     * @param args query arguments
     * @return JPQL single Result
     */
    public <T> T getSingleResultByNamedQuery(String queryName, Class<T> clazz, Object... args) {
        EntityManager em = getEntityManager();
        TypedQuery<T> q = em.createNamedQuery(queryName, clazz);
        for (int i = 0; i < args.length; i++) {
            Object arg = args[i];
            if (arg instanceof String) {
                arg = asLike((String)arg);
            }
            q.setParameter(i + 1, arg);
        }
        return q.getSingleResult();
    }

    /**
     * get/lookup object by query
     * 
     * @param query query
     * @param clazz Type of the results
     * @param <T> Type definition
     * @param args query arguments
     * @return JPQL single Result
     */
    public <T> T getSingleResultByQuery(String query, Class<T> clazz, Object... args) {
        EntityManager em = getEntityManager();
        TypedQuery<T> q = em.createQuery(query, clazz);
        for (int i = 0; i < args.length; i++) {
            Object arg = args[i];
            if (arg instanceof String) {
                arg = asLike((String)arg);
            }
            q.setParameter(i + 1, arg);
        }
        return q.getSingleResult();
    }

    /**
     * get/lookup object by query
     * 
     * @param query query
     * @param clazz Type of the results
     * @param <T> Type definition
     * @param args query arguments
     * @return JPQL single Result
     */
    public <T> T getSpecialSingleResultByQuery(String query, Class<T> clazz, Object... args) {
        EntityManager em = getEntityManager();
        TypedQuery<T> q = em.createQuery(query, clazz);
        
//        EntityGraph<T> entityGraph = em.createEntityGraph(clazz);
//        entityGraph.addAttributeNodes(graphs);
//        q.setHint("javax.persistence.fetchgraph", entityGraph);
        q.setHint("objectdb.result-fetch", "LAZY");
        for (int i = 0; i < args.length; i++) {
            Object arg = args[i];
            if (arg instanceof String) {
                arg = asLike((String)arg);
            }
            q.setParameter(i + 1, arg);
        }
        return q.getSingleResult();
    }

    /**
     * get/lookup objects by query
     * 
     * @param query query id
     * @param clazz Type of the results
     * @param <T> Type definition
     * @param args query arguments
     * @return JPQL Result list
     */
    public <T> List<T> getResultListByQuery(String query, Class<T> clazz, Object... args) {
        EntityManager em = getEntityManager();
        TypedQuery<T> q = em.createQuery(query, clazz);
        for (int i = 0; i < args.length; i++) {
            Object arg = args[i];
            if (arg instanceof String) {
                arg = asLike((String)arg);
            }
            q.setParameter(i + 1, arg);
        }
        return q.getResultList();
    }

    /**
     * execute an update by query
     * 
     * @param queryName query id
     * @param args query arguments
     */
    public void updateByNamedQuery(String queryName, Object... args) {

        EntityManager em = getEntityManager();
        Query query = em.createNamedQuery(queryName);
        query.executeUpdate();
    }

    /**
     * execute an update query
     * 
     * @param query query
     * @param args query arguments
     */
    public void updateQuery(String query, Object... args) {

        EntityManager em = getEntityManager();
        Query q = em.createQuery(query);
        q.executeUpdate();
    }

    /* *****************************************************
     * * Helper methods
     * ****************************************************
     */

    /**
     * access to entityManager
     *
     * @return EntityManager
     */
    private EntityManager getEntityManager() {
        return this.entityManager;
    }

    /**
     * map LIKE wildcards
     * 
     * @param name input String
     * @return replaced resul
     */
    private String asLike(String name) {
        String al = name.replace('*', '%');
        return al;
    }

}
